/*
 *  antidebug.c
 *
 * Fractal Guru <reverse@put.as> http://reverse.put.as
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "structures.h"

extern u_int32_t k_ptrace;
extern u_int32_t k_sysctlantidebug;
extern ptrace_func_t *real_ptrace;
extern sysctl_func_t *real_sysctl;

// Hijack ptrace syscall
// This will allow to bypass the PT_DENY_ATTACH and P_LNOATTACH (flag that denies dtrace tracing)
/*
Piece of code from ptrace @ xnu/bsd/kern/mach_process.c:

if (uap->req == PT_DENY_ATTACH) {
	proc_lock(p);
	if (ISSET(p->p_lflag, P_LTRACED)) {
		proc_unlock(p);
		exit1(p, W_EXITCODE(ENOTSUP, 0), retval);
		thread_exception_return();
	}
	SET(p->p_lflag, P_LNOATTACH);
	proc_unlock(p);
	
	return(0);
}

So if PT_DENY_ATTACH is activated then P_LNOATTACH is activated too (a local flag (p_lflag) is set)
Since we have hijacked that request, P_LNOATTACH will never be set so DTrace can work normally :).

The following simple dtrace script can be used to test this:
syscall::open*:entry
/execname == "iTunes"/
{
ustack();
}

Try to use this script with ptrace anti-anti-debug active and then disable it and try again 
(don't forget you need to quit iTunes before trying again).
Without anti-anti-debug you will get lots of errors.
*/

int our_ptrace (struct proc *p, struct ptrace_args *uap, int *retval)
{
	char processname[MAXCOMLEN+1];
	// check sysctl variable to see if we want to be active
	if (k_ptrace == 1)
	{
		if (uap->req == PT_DENY_ATTACH) {
			// grab name of the process
			proc_name(p->p_pid, processname, sizeof(processname));
			printf("[onyx-the-black-cat] Blocked PT_DENY_ATTACH/P_LNOATTACH for pid %d (%s)\n", p->p_pid, processname);
			return (0);
		} else {
			return real_ptrace(p, uap, retval);
		}
	}
	// not active so business as usual
	else
	{
		return real_ptrace(p, uap, retval);
	}
}


// our sysctl so we can intercept the anti-debug call
/*
 function that implements the sysctl syscall
 {AC(__sysctl_args), _SYSCALL_CANCEL_NONE, KERNEL_FUNNEL, (sy_call_t *)__sysctl, munge_wwwwww, munge_dddddd, _SYSCALL_RET_INT_T}, // 202 = __sysctl
 so it's "__sysctl"
 */
// __sysctl(struct proc *p, struct __sysctl_args *uap, __unused register_t *retval)
// retval is unused
// sysctl(int *name, u_int namelen, void *oldp, size_t *oldlenp, void *newp, size_t newlen);
// proc structure is defined in sys/proc_internal.h
// Hijack __sysctl
int our_sysctl (struct proc *p, struct sysctl_args *uap, int *retval)
{
	int mib[4];
	int result;
	struct kinfo_proc kpr;
	
	// call the real_sysctl function and hold the result so we can use it to return from our hijacked sysctl
	result = real_sysctl(p,uap,retval);
	
	/* 
	 **************************************************
	 How to bypass the anti-debug protection:
	 If we copy the kinfo_proc structure to kernel space and edit the p_flag and copy back to userspace,
	 then we can bypass the anti-debug protection 
	 ***************************************************
	 */
	if (k_sysctlantidebug == 1)
	{
		// int copyin(const void *uaddr, void *kaddr, size_t len); 
		// copyin() Copies len bytes of data from the user-space address uaddr to the kernel-space address kaddr.
		// copy structure from userspace to kernel space so we can verify if it's what we are looking for
		copyin(uap->name, &mib, sizeof(mib));
		// if it's a anti-debug call
		if (mib[0]==CTL_KERN && mib[1]==KERN_PROC && mib[2]==KERN_PROC_PID) 
		{
			// then copy the result from the destination buffer ( *oldp from sysctl call) to kernel space so we can edit
			copyin(uap->old,&kpr,sizeof(kpr));
			// we can display the PID from the calling program, which can be useful 
			printf("[onyx-the-black-cat] Detected possible sysctl anti-debug trick requested by process with PID %d (%s)! Patching...\n", kpr.kp_proc.p_pid, kpr.kp_proc.p_comm);
			// modify the p_flag because:
			// We're being debugged if the P_TRACED flag is set.
			// return ( (info.kp_proc.p_flag & P_TRACED) != 0 );
			kpr.kp_proc.p_flag = kpr.kp_proc.p_flag & ~P_TRACED;
			// copy back to user space the modified structure
			// int copyout(const void *kaddr, void *uaddr, size_t len);
			// copyout() Copies len bytes of data from the kernel-space address kaddr to the user-space address uaddr.
			copyout(&kpr, uap->old,sizeof(kpr));
			/* sysctl()
			 If the old value is not desired, oldp and
			 oldlenp should be set to NULL.
			 */
		}
	}
	// and return from our hijacked function
	return result;
}

