/*
 *  extendedattributes.c
 *
 * Fractal Guru <reverse@put.as> http://reverse.put.as
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "structures.h"

extern u_int32_t k_extendedattr;
extern u_int32_t k_dumpextendedattr;

extern getxattr_func_t *real_getxattr;
extern fgetxattr_func_t *real_fgetxattr;
extern setxattr_func_t *real_setxattr;
extern fsetxattr_func_t *real_fsetxattr;
extern removexattr_func_t *real_removexattr;
extern fremovexattr_func_t *real_fremovexattr;
extern listxattr_func_t *real_listxattr;
extern flistxattr_func_t *real_flistxattr;

// we can get the process name from p_comm field of a proc structure
// but it's size is limited so we don't get the full name
// we could use the proc_selfname or proc_name name functions but they aren't any better since they retrieve the same p_comm field
// Hijack getxattr
int our_getxattr (struct proc *p, struct getxattr_args *uap, int *retval)
{
	int result, i=0, error;
	size_t attributenamelen, attributesize;
	char attributename[XATTR_MAXNAMELEN+1];
	char processname[MAXCOMLEN+1];
	
	attributesize = uap->size;
	char attributevalue[attributesize];
	// handle request for attribute size (value equal to NULL)
	if (uap->value == 0)
	{ 
		result = real_getxattr(p, uap, retval);
	}
	else if (k_extendedattr == 1)
	{
		if (error = copyinstr(uap->attrname, attributename, sizeof(attributename), &attributenamelen) != 0)
		{
			printf("[onyx-the-black-cat] Error: Can't copy attribute name to kernel space!\n");
		}
		/* from sys/proc.h
		 this routine copies the process's name of the executable to the passed in buffer. It
		 is always null terminated. The size of the buffer is to be passed in as well. This
		 routine is to be used typically for debugging
		 void proc_name(int pid, char * buf, int size);
		 
		 FIXME: In tiger accessing p->p_comm gives us the name but in Leopard doesnt...
		 proc_name must be used, strange since this function just copies from the same proc structure... To be analysed or not :)
		 */
		proc_name(p->p_pid, processname, sizeof(processname));
		printf("[onyx-the-black-cat] Detected request to getxattr syscall from PID %d (%s) Attribute name: %s Value: ",p->p_pid, processname, attributename);
		result = real_getxattr(p, uap, retval);
		copyin(uap->value, &attributevalue, sizeof(attributevalue));
		printf("%s\n",attributevalue);
		if (k_dumpextendedattr == 1)
		{
			printf("[onyx-the-black-cat] getxattr Hex dump: ");
			for (i=0; i<sizeof(attributevalue); i++) { printf("%x",attributevalue[i]); }
			printf("\n");
		}
	}
	else
	{
		result = real_getxattr(p, uap, retval);
	}
	return result;
}

// Hijack fgetxattr
// kernel implementation at xnu/bsd/vfs/vfs_syscalls.c
int our_fgetxattr (struct proc *p, struct fgetxattr_args *uap, int *retval)
{
	int result, i=0, error;
	size_t attributenamelen, attributesize;
	char attributename[XATTR_MAXNAMELEN+1];
	char processname[MAXCOMLEN+1];
	
	attributesize = uap->size;
	char attributevalue[attributesize];
	/* man getxattr:
     On success, value contains the data associated with name.  When value is
     set to NULL, getxattr() returns current size of the named attribute.
     This facility can be used to determine the size of a buffer sufficiently
     large to hold the data currently associated with the attribute.
	 */
	// we are just interested in dumping real reads to the attribute name
	if (uap->value == 0)
	{
		result = real_fgetxattr(p, uap, retval);
	}
	else if (k_extendedattr == 1)
	{
		// copy attribute name to kernel space
		if(error = copyinstr(uap->attrname, attributename, sizeof(attributename), &attributenamelen) != 0)
		{
			printf("[onyx-the-black-cat] Error: Can't copy attribute name to kernel space!\n");
		}
		// grab process name
		proc_name(p->p_pid, processname, sizeof(processname));
		printf("[onyx-the-black-cat] Detected request to fgetxattr syscall from PID %d (%s) Attribute name: %s Value: ", p->p_pid, p->p_comm, attributename);
		// call the real fgetxattr so we can have value
		result = real_fgetxattr(p, uap, retval);
		// copy value to kernel space
		copyin(uap->value, &attributevalue, sizeof(attributevalue));
		printf("%s\n", attributevalue);
		if (k_dumpextendedattr == 1)
		{
			printf("[onyx-the-black-cat] fgetxattr Hex dump: ");
			for (i=0; i<sizeof(attributevalue); i++) { printf("%x",attributevalue[i]); }
			printf("\n");
		}
	}
	else
	{
		result = real_fgetxattr(p, uap, retval);
	}
	return result;
}

// Hijack setxattr
int our_setxattr (struct proc *p, struct setxattr_args *uap, int *retval)
{
	int result;
	char attributepath[255];
	char attributename[XATTR_MAXNAMELEN+1];
	char processname[MAXCOMLEN+1];
	if (k_extendedattr == 1)
	{
		proc_name(p->p_pid, processname, sizeof(processname));
		copyin(uap->path, &attributepath, sizeof(attributepath));
		copyin(uap->attrname, &attributename, sizeof(attributename));
		printf("[onyx-the-black-cat] Detected request to setxattr syscall from PID %d (%s) Path: %s Attribute name: %s\n", p->p_pid, processname, attributepath, attributename);
		result = real_setxattr(p, uap, retval);
	}
	else
	{
		result = real_setxattr(p, uap, retval);
	}
	return result;
}

// Hijack fsetxattr
int our_fsetxattr (struct proc *p, struct fsetxattr_args *uap, int *retval)
{
	int result;
	char processname[MAXCOMLEN+1];
	if (k_extendedattr == 1)
	{
		proc_name(p->p_pid, processname, sizeof(processname));
		printf("[onyx-the-black-cat] Detected request to fsetxattr syscall from PID %d (%s)\n", p->p_pid, processname);
		result = real_fsetxattr(p, uap, retval);
	}
	else
	{
		result = real_fsetxattr(p, uap, retval);
	}
	return result;
}

// Hijack removexattr
int our_removexattr (struct proc *p, struct removexattr_args *uap, int *retval)
 {
	 int result;
	 char attributepath[255];
	 char attributename[XATTR_MAXNAMELEN+1];
	 char processname[MAXCOMLEN+1];
	 if (k_extendedattr == 1)
	 {
		 proc_name(p->p_pid, processname, sizeof(processname));
		 copyin(uap->path, &attributepath, sizeof(attributepath));
		 copyin(uap->attrname, &attributename, sizeof(attributename));
		 printf("[onyx-the-black-cat] Detected request to removexattr syscall from PID %d (%s) Path: %s Attribute name: %s\n", p->p_pid, processname, attributepath, attributename);
		 result = real_removexattr(p, uap, retval);
	 }
	 else
	 {
		 result = real_removexattr(p, uap, retval);
	 }
	 return result;
 }

// Hijack fremovexattr
int our_fremovexattr (struct proc *p, struct fremovexattr_args *uap, int *retval)
{
	int result;
	char processname[MAXCOMLEN+1];
	char attributename[XATTR_MAXNAMELEN+1];
	
	if (k_extendedattr == 1)
	{
		proc_name(p->p_pid, processname, sizeof(processname));
		copyin(uap->attrname, &attributename, sizeof(attributename));
		printf("[onyx-the-black-cat] Detected request to fremovexattr syscall from PID %d (%s) Attribute name: %s\n", p->p_pid, processname, attributename);
		result = real_fremovexattr(p, uap, retval);
	}
	else
	{
		result = real_fremovexattr(p, uap, retval);
	}
	return result;
}

// Hijack listxattr
int our_listxattr (struct proc *p, struct listxattr_args *uap, int *retval)
{
	int result;
	char attributepath[255];
	char processname[MAXCOMLEN+1];
	/*
     If namebuf is set to NULL, the function returns the size of the list of
     extended attribute names.  This facility can be used to determine the
     size of a buffer sufficiently large to hold the names of the attributes
     currently associated with path.
	 */
	if (uap->namebuf == 0)
	{
		result = real_listxattr(p, uap, retval);
	}
	else if (k_extendedattr == 1)
	{
		proc_name(p->p_pid, processname, sizeof(processname));
		copyin(uap->path, &attributepath, sizeof(attributepath));
		printf("[onyx-the-black-cat] Detected request to listxattr syscall from PID %d (%s) Path: %s\n", p->p_pid, processname, attributepath);
		result = real_listxattr(p, uap, retval);
	}
	else
	{
		result = real_listxattr(p, uap, retval);
	}
	return result;
}

// Hijack flistxattr
int our_flistxattr (struct proc *p, struct flistxattr_args *uap, int *retval)
{
	int result;
	char processname[MAXCOMLEN+1];
	
	if (uap->namebuf == 0)
	{
		result = real_flistxattr(p, uap, retval);
	}
	else if (k_extendedattr == 1)
	{
		proc_name(p->p_pid, processname, sizeof(processname));
		printf("[onyx-the-black-cat] Detected request to flistxattr syscall from PID %d (%s)\n", p->p_pid, processname);
		result = real_flistxattr(p, uap, retval);
	}
	else
	{
		result = real_flistxattr(p, uap, retval);
	}
	return result;
}

