/*
 * onyx-the-black-cat.c
 * Fractal Guru <reverse@put.as> http://reverse.put.as
 *
 * For use with Mac OS X Leopard 10.5.6
 *
 * Based on original code by Landon J. Fuller <landonf@opendarwin.org>
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#define VERSION 0.3

/* COMPILE OPTIONS */
// change to 0 if to remove monitoring of extended attributes calls
#define MONITOR_EXTENDED_ATTRIBUTES 1

#include "structures.h"

static struct sysent *_sysent;
extern int nsysent;

// reserve space to store the original functions
// *_func_t definition @ structures.h
ptrace_func_t *real_ptrace;
sysctl_func_t *real_sysctl;
getxattr_func_t *real_getxattr;
fgetxattr_func_t *real_fgetxattr;
setxattr_func_t *real_setxattr;
fsetxattr_func_t *real_fsetxattr;
removexattr_func_t *real_removexattr;
fremovexattr_func_t *real_fremovexattr;
listxattr_func_t *real_listxattr;
flistxattr_func_t *real_flistxattr;
open_func_t *real_open;
chmod_func_t *real_chmod;

/* sysctl variables support */
u_int32_t k_ptrace = 1;        // the contents of onyx.activate
u_int32_t k_sysctlantidebug = 1;        // the contents of onyx.sysctltrick
u_int32_t k_extendedattr = 0;	// by default do not monitor extended attributes
u_int32_t k_dumpextendedattr = 0; // by default do not hex dump content of extended attributes (if monitoring is active)

// Construct a node (debug.onyx) from which other sysctl objects can hang.
SYSCTL_NODE(_debug,       // our parent
            OID_AUTO,     // automatically assign us an object ID
            onyx,		  // our name
            CTLFLAG_RW,   // we will be creating children, therefore, read/write
            0,            // handler function (none needed)
            "onyx the black cat sysctl hierarchy");

// Prototypes for read/write handling functions for our sysctl nodes.
static int sysctl_onyx_ptrace SYSCTL_HANDLER_ARGS;
static int sysctl_onyx_sysctlantidebug SYSCTL_HANDLER_ARGS;
static int sysctl_onyx_extendedattr SYSCTL_HANDLER_ARGS;
static int sysctl_onyx_dumpextendedattr SYSCTL_HANDLER_ARGS;

// We can directly use SYSCTL_INT(), in which case sysctl_handle_int()
// will be assigned as the handling function. We use SYSCTL_PROC() and
// specify our own handler sysctl_osxbook_uint32().
SYSCTL_PROC(
			_debug_onyx,                 // our parent
			OID_AUTO,                       // automatically assign us an object ID
			ptrace,                         // our name
			(CTLTYPE_INT |                  // type flag (available at xnu/bsd/sys/sysctl.h)
			 CTLFLAG_RW), // access flags (read/write by root only - if we want by everybody add CTLFLAG_ANYBODY flag)
			&k_ptrace,                      // location of our data
			0,                              // argument passed to our handler
			sysctl_onyx_ptrace,          // our handler function
			"IU",                           // our data type (unsigned integer)
			"PT_DENY_ATTACH anti-debug"       // our description
);
SYSCTL_PROC(_debug_onyx,OID_AUTO,sysctlantidebug,(CTLTYPE_INT | CTLFLAG_RW), &k_sysctlantidebug,0,sysctl_onyx_sysctlantidebug,"IU", "sysctl anti-debug");
SYSCTL_PROC(_debug_onyx,OID_AUTO,extendedattr,(CTLTYPE_INT | CTLFLAG_RW), &k_extendedattr,0,sysctl_onyx_extendedattr,"IU", "monitor extended attributes");
SYSCTL_PROC(_debug_onyx,OID_AUTO,dumpextendedattr,(CTLTYPE_INT | CTLFLAG_RW), &k_dumpextendedattr,0,sysctl_onyx_dumpextendedattr,"IU", "hex dump extended attributes");

static int sysctl_onyx_ptrace SYSCTL_HANDLER_ARGS
{
    // Do some processing of our own, if necessary.
    return sysctl_handle_int(oidp, oidp->oid_arg1, oidp->oid_arg2, req);
}

static int sysctl_onyx_sysctlantidebug SYSCTL_HANDLER_ARGS
{
    // Do some processing of our own, if necessary.
    return sysctl_handle_int(oidp, oidp->oid_arg1, oidp->oid_arg2, req);
}

static int sysctl_onyx_extendedattr SYSCTL_HANDLER_ARGS
{
    // Do some processing of our own, if necessary.
    return sysctl_handle_int(oidp, oidp->oid_arg1, oidp->oid_arg2, req);
}

static int sysctl_onyx_dumpextendedattr SYSCTL_HANDLER_ARGS
{
    // Do some processing of our own, if necessary.
    return sysctl_handle_int(oidp, oidp->oid_arg1, oidp->oid_arg2, req);
}

/*
 * nsysent is placed directly before the hidden sysent, so skip ahead
 * and sanity check that we've found the sysent array.
 *
 * Clearly, this is extremely fragile and not for general consumption.
 */
static struct sysent *find_sysent () 
{
	unsigned int table_size;
	struct sysent *table;

	table_size = sizeof(struct sysent) * nsysent;
	table = (struct sysent *) ( ((char *) &nsysent) + sizeof(nsysent) );

#if __i386__
	/* For reasons unknown, the table is offset by an additional 28 bytes on my i386 system */
	table = (struct sysent *) ( ((uint8_t *) table) + 28);
#endif

	printf("[onyx-the-black-cat] Found nsysent at %p (count %d), calculated sysent location at %p.\n", &nsysent, nsysent, table);

	/* Sanity check */
	printf("[onyx-the-black-cat] Sanity check: verifying if number of syscalls arguments are the expected ones\n");
		
// number of arguments for each function (we can use xnu/bsd/sys/sysproto.h)
// if sysent table is found and it's correct, the number of arguments should match
	if (
		table[SYS_exit].sy_narg == 1 &&
		table[SYS_fork].sy_narg == 0 &&
		table[SYS_read].sy_narg == 3 &&
		table[SYS_wait4].sy_narg == 4 &&
		table[SYS_ptrace].sy_narg == 4 &&
		table[SYS_getxattr].sy_narg == 6 &&
		table[SYS_listxattr].sy_narg == 4)
	{
		printf("[onyx-the-black-cat] Sanity check: sysent sanity check succeeded.\n");
		return table;
	} else {
		printf("[onyx-the-black-cat] Sanity check: sanity check failed, could not find sysent table.\n");
		return NULL;
	}
}

// THE FUN STARTS HERE
kern_return_t onyx_the_black_cat_start (kmod_info_t * ki, void * d) {
    // Register our sysctl entries.
	sysctl_register_oid(&sysctl__debug_onyx);
    sysctl_register_oid(&sysctl__debug_onyx_ptrace);
    sysctl_register_oid(&sysctl__debug_onyx_sysctlantidebug);
	sysctl_register_oid(&sysctl__debug_onyx_extendedattr);
	sysctl_register_oid(&sysctl__debug_onyx_dumpextendedattr);
	
	printf("[onyx-the-black-cat] Starting patching ...\n");
    printf("[onyx-the-black-cat] Finding sysent table...\n");
	// get sysent address
	_sysent = find_sysent();
	if (_sysent == NULL) {
	    printf("[onyx-the-black-cat] Error: Cannot find sysent table\n");
		return KERN_FAILURE;
	}
	
	// save address for the real functions
	real_ptrace = (ptrace_func_t *) _sysent[SYS_ptrace].sy_call;
	real_sysctl = (sysctl_func_t *) _sysent[SYS___sysctl].sy_call;
#if MONITOR_EXTENDED_ATTRIBUTES >= 1
	real_getxattr = (getxattr_func_t *) _sysent[SYS_getxattr].sy_call;
	real_fgetxattr = (fgetxattr_func_t *) _sysent[SYS_fgetxattr].sy_call;
	real_setxattr = (setxattr_func_t *) _sysent[SYS_setxattr].sy_call;
	real_fsetxattr = (fsetxattr_func_t *) _sysent[SYS_fsetxattr].sy_call;
	real_removexattr = (removexattr_func_t *) _sysent[SYS_removexattr].sy_call;
	real_fremovexattr = (fremovexattr_func_t *) _sysent[SYS_fremovexattr].sy_call;
	real_listxattr = (listxattr_func_t *) _sysent[SYS_listxattr].sy_call;
	real_flistxattr = (flistxattr_func_t *) _sysent[SYS_flistxattr].sy_call;
#endif
//	real_chmod = (chmod_func_t *)_sysent[SYS_chmod].sy_call;
//	real_open = (open_func_t *) _sysent[SYS_open].sy_call;
	
	// replace the original functions with ours, aka HIJACK
	_sysent[SYS_ptrace].sy_call = (sy_call_t *) our_ptrace;
	_sysent[SYS___sysctl].sy_call = (sy_call_t *) our_sysctl;
#if MONITOR_EXTENDED_ATTRIBUTES >= 1
	_sysent[SYS_getxattr].sy_call = (sy_call_t *) our_getxattr;
	_sysent[SYS_fgetxattr].sy_call = (sy_call_t *) our_fgetxattr;
	_sysent[SYS_setxattr].sy_call = (sy_call_t *) our_setxattr;
	_sysent[SYS_fsetxattr].sy_call = (sy_call_t *) our_fsetxattr;
	_sysent[SYS_removexattr].sy_call = (sy_call_t *) our_removexattr;
	_sysent[SYS_fremovexattr].sy_call = (sy_call_t *) our_fremovexattr;
	_sysent[SYS_listxattr].sy_call = (sy_call_t *) our_listxattr;
	_sysent[SYS_flistxattr].sy_call = (sy_call_t *) our_flistxattr;
#endif
//	_sysent[SYS_chmod].sy_call = (sy_call_t *) our_chmod;
//	_sysent[SYS_open].sy_call = (sy_call_t *) our_open;
	
	printf("[onyx-the-black-cat] Patching is successful! Have fun...\n");
	// ALL DONE
	return KERN_SUCCESS;
}


// THE FUN ENDS HERE
kern_return_t onyx_the_black_cat_stop (kmod_info_t * ki, void * d) {
	
	printf("[onyx-the-black-cat] Starting unpatching ...\n");

    _sysent[SYS_ptrace].sy_call = (sy_call_t *) real_ptrace;
	_sysent[SYS___sysctl].sy_call = (sy_call_t *) real_sysctl;
#if MONITOR_EXTENDED_ATTRIBUTES >= 1
	_sysent[SYS_getxattr].sy_call = (sy_call_t *) real_getxattr;
	_sysent[SYS_fgetxattr].sy_call = (sy_call_t *) real_fgetxattr;
	_sysent[SYS_setxattr].sy_call = (sy_call_t *) real_setxattr;
	_sysent[SYS_fsetxattr].sy_call = (sy_call_t *) real_fsetxattr;
	_sysent[SYS_removexattr].sy_call = (sy_call_t *) real_removexattr;
	_sysent[SYS_fremovexattr].sy_call = (sy_call_t *) real_fremovexattr;
	_sysent[SYS_listxattr].sy_call = (sy_call_t *) real_listxattr;
	_sysent[SYS_flistxattr].sy_call = (sy_call_t *) real_flistxattr;
#endif
//	_sysent[SYS_open].sy_call = (sy_call_t *) real_open;
//	_sysent[SYS_chmod].sy_call = (sy_call_t *) real_chmod;

	printf("[onyx-the-black-cat] Unpatching successful ...\n");
	
	// Unregister our sysctl entries.
    sysctl_unregister_oid(&sysctl__debug_onyx_ptrace);
    sysctl_unregister_oid(&sysctl__debug_onyx_sysctlantidebug);
	sysctl_unregister_oid(&sysctl__debug_onyx_extendedattr);
	sysctl_unregister_oid(&sysctl__debug_onyx_dumpextendedattr);
    sysctl_unregister_oid(&sysctl__debug_onyx);
	
    // ALL DONE
	return KERN_SUCCESS;
}
